import os
import base64
import hashlib
from ch35t.utils import text_decode
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes

class NullOutputDirException(Exception):
    pass

class MethodHandler:
    '''
        Handles payloads (depending on encryption method, it
        applies different algorithms to unlock/lock a chest).
        Identified by its method (no two handlers
        have the same one).
    '''
    method = None

    def __init__(self, data, ctx):
        if ctx is None or ctx.get('output_dir', None) is None:
            raise NullOutputDirException

        # simply store data for later use
        self._data = data
        self._ctx = ctx
        self._output_dir = ctx['output_dir']
        self._cleartext_data = None

    def lock(self, key):
        '''
            Applies a cryptographic algorithm to lock self._data with the
            provided key.
            Regardless of the employed algorithm, locking should always
            return the encrypted data as a string (possibly base64-encoded)
        '''
        raise NotImplementedError

    def unlock(self, key):
        '''
            Applies a cryptographic algorithm to unlock self._data with the
            provided key.
            "Unlocking" has different outcomes depending on the algorithm
            (e.g. the key can be verified against a hash, used to decrypt 
            some content, etc.).

            After unlocking, this function does the following:
            - update self._cleartext_data if any data is generated
              (dealing with cleartext data is a task for payload class)
            - return True if the key was correct, False otherwise
        '''
        raise NotImplementedError


class MD5Handler(MethodHandler):
    method = "md5"

    def lock(self):
        '''
            Returns the MD5 hash of self._data
        '''

        return hashlib.md5(self._data.encode('utf-8')).hexdigest()

    def unlock(self, key):
        '''
            Checks the key against a provided MD5 hash. Returns
            True if the hashed key matches, False otherwise
        '''

        # base64 is probably never used together with MD5...
        # but who knows :-)
        self._data = text_decode(self._data)
        if hashlib.md5(key.encode('utf-8')).hexdigest() == self._data:
            return True
        else:
            return False


class AESHandler(MethodHandler):
    method = "AES"

    def _get_cipher(self, key, iv=None):
        '''
            Given a key and an optional initialization vector,
            returns an AES cipher object that can be used
            both for encryption and decryption.

            For encryption, it is suggested to always use random
            ivs (if none is passed to the function, a random one
            will be created)
        '''

        # to have consistent, 32-bytes key length,
        # the key is converted into its md5 hash
        key = bytes(hashlib.md5(bytes(key, 'utf-8')).hexdigest(), 'utf-8')

        if iv is None:
            # set the initialization vector to 16 random bytes
            iv = os.urandom(16)

        # create an AES encryptor object
        cipher = Cipher(algorithms.AES(key), modes.CBC(iv))

        return cipher, iv

    def lock(self, key):
        '''
            Given a key, encrypts data using the AES algorithm.
            Data is returned as base64-encoded so it can be easily
            handled in ch35t json files.
        '''
        cipher, iv = self._get_cipher(key)
        encryptor = cipher.encryptor()

        # encrypt
        ct = encryptor.update(bytes(self._data, 'utf-8')) + encryptor.finalize()

        # concatenate iv and ct
        full_message = bytearray(iv)
        full_message.extend(ct)
        full_message = base64.b64encode(full_message).decode('utf-8')

        # update self._data
        return f"base64,{full_message}"

    def unlock(self, key):
        '''
            Given a key, decrypts data using the AES algorithm.
            If decrypted data is not junk (i.e. utf-8 decoding
            does not break), it returns True and saves it as
            cleartext_data, otherwise it returns False
        '''

        # data is provided as base64-encoded, so decode it first
        data = text_decode(self._data)

        # recover iv and ct from data
        iv = data[:16]
        ct = data[16:]

        cipher,_ = self._get_cipher(key, iv)
        decryptor = cipher.decryptor()

        # decrypt
        pt = decryptor.update(ct) + decryptor.finalize()

        # we have no real way to know whether the key is correct
        # as AES will decrypt in any case, but if what is returned
        # is garbage the utf-8 decoding will break
        try:
            pt = pt.decode('utf-8')
        except UnicodeDecodeError as e:
            return False

        self._cleartext_data = pt
        return True


# Dictionary mapping mime types to classes
handlers = {cls.method: cls for cls in MethodHandler.__subclasses__()}

def get_handler(method, data, ctx):
    if method in handlers:
        return handlers[method](data, ctx)
    else:
        raise ValueError(f"Unsupported method: {method}")



from ch35t.hint import Hint
from ch35t.payload import Payload
from ch35t.label import Label
import json
import requests


class ParserException(Exception):
        def __init__(self, message):
                    super().__init__(message)


class ChestLoader():
    ### TODO: add better error handling here
    def from_file(fname):
        try:
            with open(fname, "rt") as f:
                return json.load(f)
        except Exception as e:
            raise ParserException(e)
        
    def from_string(string):
        try:
            return json.loads(string)
        except Exception as e:
            raise ParserException(e)
    
    def from_url(url):
        try:
            res = requests.get(url)
            if res.status_code == 200:
                return json.loads(res.text)
            return None
        except Exception as e:
            raise ParserException(e)


class Parser():

    def __init__(self, string=None):
        
        self._json = None
        self._hint = None
        self._label = None
        self._payload = None
        self._method = None

        # If no string is provided as input, then the local json is uninitialized.
        # Note that it is possible to do so and manually load the json by calling 
        # one of the self._load_* methods defined before (dunno why yet, but I am
        # confident it might come useful at some point :-)
        if string is not None:
            if string.startswith("http://") or string.startswith("https://"):
                self._load_url(string)
            elif string.startswith("file://"):
                self._load_file(string[7:])
            else:
                self._load_string(string)
                
        # JSON is parsed only if not null
        # This is done to allow for later manual loading of JSON,
        # or to take JSON parsing errors into account
        if self._json is not None:
            self._parse_json()
    
    def _load_string(self, string):
        self._json = ChestLoader.from_string(string)

    def _load_file(self, fname):
        self._json = ChestLoader.from_file(fname)
    
    def _load_url(self, url):
        self._json = ChestLoader.from_url(url)

    def _parse_json(self):
        self._hint = Hint(self._json)
        self._payload = Payload(self._json)
        self._label = Label(self._json)

    def json(self):
        return self._json

    def hint(self):
        return self._hint

    def payload(self):
        return self._payload

    def label(self):
        return self._label

### TODO: add some main code here showing how a parser works



import click
import click_repl
from ch35t.ch35t import Chest
import os
from prompt_toolkit.history import FileHistory


def custom_help(ctx, sections=['usage', 'short_help_str', 'options', 'commands'], return_dict=False):
    """ 
        Return only a part of the sections the default help is split into

        click's help is great, but it assumes every command is always ran
        directly from the command line. Being in a CLI, the default help
        is a bit misleading.

        This method allows one to choose which sections of the default help
        are shown, so e.g. one can skip "usage" which is thought for the CLI.

        Using this help is still a bit cumbersome as it requires to disable
        the default one and explicitly call this method, so I am actually
        not sure if/how much I am going to use it... Hey, this is what you
        get for trying some highly-under-development code :-)
    """

    help_str = ctx.get_help()
    help_dict = {}
    help_sections = ['usage', 'short_help_str', 'options', 'commands']
    idx = 0

    # clean sections from commands if not group:
    if type(ctx.command) != click.core.Group and 'commands' in sections:
        sections.remove('commands')

    # init dict
    for k in help_sections:
        help_dict[k] = ""

    # parse help and split it in different sections
    for line in help_str.split("\n"):
        if idx == 0:
            script_name = " "+__file__.split("/")[-1]
            # remove script name from usage
            line = line.replace(script_name,"")
        if line == "":
            idx += 1
            continue
        else:
            help_dict[help_sections[idx]] += f"{line}\n"

    if return_dict:
        return(help_dict)

    help_out = ""
    for section in sections:
        help_out += help_dict[section] + "\n"
    
    return(help_out)
    

@click.group(invoke_without_command=True)
@click.pass_context
def cli(ctx):
    """CLI-based ch35t manipulation tool"""

    # initialize the obj (a dict) that will hold our chest
    # (and possibly other metadata we want to persist inside the REPL)
    ctx.ensure_object(dict)

    if ctx.invoked_subcommand is None:
        ctx.invoke(repl)


@cli.command()
@click.pass_context
def status(ctx):
    """Shows the current status"""
    
    click.secho("You look around you and you see...", bold=True)

    chest = ctx.obj.get("chest")
    if chest is None:
        click.echo("🤷 no chest is around, go get one!")
    else:
        click.echo("📦 you have a chest with you")
        click.echo(f" - the chest has a label on it, saying its name is ", nl=False)
        click.secho(chest.name(), bold=True)
        if chest.author() is not None:
            click.echo(f" - there is also a name on the label, it appears the chest creator is ", nl=False)
            click.secho(chest.author(), bold=True)
        else:
            click.echo(f" - the creator of this chest is unknown")

    click.echo()


@cli.command()
@click.pass_context
@click.option('--json', help="Chest JSON (string, file:// or http(s)://)")
def get(ctx, json):
    """Gets a chest (JSON string, file, or URL)"""

    if json is None:
        # keep init with null JSON disabled until we have all the
        # methods for manual loading exposed
        click.secho("You should provide a chest in JSON format", fg='yellow')
    else:
        try:
            ctx.obj["chest"] = Chest(json, chests_dir="./chests")
            click.secho("📦 I got a chest! Now check if it is a valid one", fg='green')
        except Exception as e:
            click.secho(f"[x] {e}", fg='red')


@cli.command()
@click.pass_context
@click.option('--dump', help="Just show the raw hint object", is_flag=True, default=False, show_default=True)
def hint(ctx, dump):
    """Shows a chest's hint"""
    chest = ctx.obj.get("chest")
    if chest is None:
        click.secho("It appears you have no chest, try to get one first", fg='yellow')
    else:
        if dump:
            click.echo(chest.hint.dump())
        else:
            click.secho("Here is the hint to open your chest:", bold=True)
            click.echo(chest.hint)


@cli.command()
@click.pass_context
@click.option('--dump', help="Just show the raw payload object", is_flag=True, default=False, show_default=True)
def payload(ctx, dump):
    """Shows a chest's payload"""
    chest = ctx.obj.get("chest")
    if chest is None:
        click.secho("It appears you have no chest, try to get one first", fg='yellow')
    else:
        if dump:
            click.echo(chest.payload.dump())
        else:
            click.secho("Here is the content of your chest:", bold=True)
            click.echo(chest.payload)


@cli.command()
@click.pass_context
@click.option('--dump', help="Just show the raw hint object", is_flag=True, default=False, show_default=True)
def label(ctx, dump):
    """Shows a chest's label"""
    chest = ctx.obj.get("chest")
    if chest is None:
        click.secho("It appears you have no chest, try to get one first", fg='yellow')
    else:
        if dump:
            click.echo(chest.label.dump())
        else:
            click.secho("You look at the label attached to the chest and read the following:", bold=True)
            click.echo("-------------------------------------------------------------------")
            click.echo("The name of this chest is: ", nl=False)
            click.secho(chest.name(), bold=True)
            click.echo("The name of its creator is: ", nl=False)
            author = "Unknown" if chest.author() is None else chest.author()
            click.secho(author, bold=True)


@cli.command(add_help_option=False)
@click.pass_context
def help(ctx):
    """Well you are here, so I suppose you know already ;-)"""
    # get custom help dictionary  
    help_dict = custom_help(ctx.parent, return_dict=True)

    click.echo()
    click.secho(help_dict['short_help_str'], bold=True, nl=False)
    click.echo("------------------------------------------\n")
    click.echo(help_dict['commands'])
    click.echo("Type \"COMMAND --help\" to get help about a command\n")


@cli.command()
@click.pass_context
def validate(ctx):
    """Validate the chest JSON according to Ch35t schema"""
    chest = ctx.obj.get("chest")
    if chest is None:
        click.secho("It appears you have no chest, try to get one first", fg='yellow')
    else:
        try:
            chest.validate()
            click.secho("This chest is valid!", fg='green')
        except Exception as e:
            click.secho("Oh no, there seems to be a validation error:", fg='red')
            print(e)


@cli.command()
@click.pass_context
@click.option("--key", prompt="Enter the key to open the chest", help="The key to be used to open the chest")
def unlock(ctx, key):
    """Unlock a chest provided a key"""
    chest = ctx.obj.get("chest")
    if chest is None:
        click.secho("It appears you have no chest, try to get one first", fg='yellow')
    else:
        if chest.unlock(key):
            click.secho("The chest is now unlocked!", fg='green')
        else:
            click.secho("The key is not the right one, the chest is still locked", fg='red')


def repl():
    """Start an interactive session"""
    prompt_kwargs = {
        'history': FileHistory(os.path.expanduser('~/.repl_history'))
    }
    click_repl.repl(click.get_current_context(), prompt_kwargs=prompt_kwargs)


def main():
    cli(obj={})

if __name__ == '__main__':
    main()


import zipfile
import io
from ch35t.utils import text_decode

# application/octet-stream is used for generic binary files

class NullOutputDirException(Exception):
    pass

class DataHandler:
    '''
        Handles hints/payloads data so it can be easily accessed.
        Identified by its mime_type (no two handlers have the
        same one).
    '''
    mime_type = None

    def __init__(self, data, ctx):
        if ctx is None or ctx.get('output_dir', None) is None:
            raise NullOutputDirException

        # simply store text-decoded data for later use
        self._data = text_decode(data)
        self._ctx = ctx
        self._output_dir = ctx['output_dir']

    def to_string(self):
        '''
            Returns data (or a textual description of it) as a string
        '''
        return self._data

    def handle(self):
        '''
            
            (returns None)
        '''
        raise NotImplementedError


class TextDataHandler(DataHandler):
    mime_type = "text/plain"

    def __init__(self, data, ctx):
        super().__init__(data, ctx)

    def handle(self):
        # if data was base64-encoded, and after decoding it we have bytes,
        # we can assume text was encoded (e.g. in utf-8)
        # TODO: add support for different encodings, for now we only
        #       consider utf-8
        if type(self._data) == bytes:
            self._data = self._data.decode('utf-8')


class ZipDataHandler(DataHandler):
    mime_type = "application/zip"

    def __init__(self, data, ctx):
        super().__init__(data, ctx)

    def handle(self):
        with zipfile.ZipFile(io.BytesIO(self._data), "r") as zf:
            zf.extractall(self._output_dir)

    def to_string(self):
        s = f"Files are available in {self._output_dir}:\n"
        with zipfile.ZipFile(io.BytesIO(self._data), "r") as zf:
            for fileinfo in zf.infolist():
                s+=f"{fileinfo.filename}\n"
        return s

# Dictionary mapping mime types to classes
handlers = {cls.mime_type: cls for cls in DataHandler.__subclasses__()}

def get_handler(mime_type, data, ctx):
    if mime_type in handlers:
        return handlers[mime_type](data, ctx)
    else:
        raise ValueError(f"Unsupported mime type: {mime_type}")



import ch35t.methodhandlers as mh
import ch35t.datahandlers as dh

class NullPayloadException(Exception):
    def __init__(self, message):
        super().__init__(message)

class Payload():
    _default_format = "text/plain"

    def __init__(self, json_data):
        payload = json_data.get('payload')
        if payload is None or (payload.get('data') is None and payload.get('origin') is None):
            raise NullPayloadException("No payload has been provided in this chest's JSON")

        self.origin = payload.get('origin')
        self.data = payload.get('data')
        self.fmt = payload.get('format')
        self.cleartext_data = None
        self._methodhandler = None
        self._datahandler = None

        # if a method is specified (note that it is optional!), get its name
        self.method = None
        m = payload.get('method')
        if m is not None:
            self.method = m.get('name')
 
        if self.fmt is None:
            self.fmt = self._default_format

    def set_ctx(self, ctx):
        self._ctx = ctx

    def unlock(self, key):
        # calls methodhandler to unlock the payload
        if self._methodhandler is None:
            self._methodhandler = mh.get_handler(self.method, self.data, self._ctx)

        if self._methodhandler.unlock(key):
            self.cleartext_data = self._methodhandler._cleartext_data
            return True

        return False

    def __str__(self):
        if self._datahandler is None:
            if not self.cleartext_data:
                ctdata = self.data

            if hasattr(self, "_ctx"):
                context = self._ctx
            else:
                context = {"output_dir": "/tmp"}

            self._datahandler = dh.get_handler(self.fmt, ctdata, context)
            self._datahandler.handle()

        return self._datahandler.to_string()

    def dump(self):
        return_str  =  "[i] Payload\n"
        return_str += f"    Origin: {self.origin}\n"
        return_str += f"    Data: {self.data}\n"
        return_str += f"    Cleartext Data: {self.cleartext_data}\n"
        return_str += f"    Format: {self.fmt}\n"
        return_str += f"    Method: {self.method}\n"
        return return_str
    



from ch35t.datahandlers import get_handler

class NullHintException(Exception):
    def __init__(self, message):
        super().__init__(message)

class Hint():
    _default_format = "text/plain"

    def __init__(self, json_data):
        hint = json_data.get('hint')
        if hint is None or (hint.get('data') is None and hint.get('origin') is None):
            raise NullHintException("No hint has been provided in this chest's JSON")

        self.origin = hint.get('origin')
        self.data = hint.get('data')
        self.fmt = hint.get('format')
        if self.fmt is None:
            self.fmt = self._default_format
        self._datahandler = None

    def set_ctx(self, ctx):
        self._ctx = ctx

    def __str__(self):
        if self._datahandler is None:
            self._datahandler = get_handler(self.fmt, self.data, self._ctx)
            self._datahandler.handle()
        return self._datahandler.to_string()


    def dump(self):
        return_str  =  "[i] Hint\n"
        return_str += f"    Origin: {self.origin}\n"
        return_str += f"    Data: {self.data}\n"
        return_str += f"    Format: {self.fmt}\n"
        return return_str



class NullLabelException(Exception):
    def __init__(self, message):
        super().__init__(message)

class Label():

    def __init__(self, json_data):
        label = json_data.get('label')
        if label is None or label.get('name') is None:
            raise NullLabelException("No valid label has been provided in this chest's JSON")

        self.name = label.get('name')
        self.author = label.get('author')
        self.URI = label.get('URI')


    def dump(self):
        return_str  =  "[i] Label\n"
        return_str += f"    Name: {self.name}\n"
        return_str += f"    Author: {self.author}\n"
        return_str += f"    URI: {self.URI}\n"
        return return_str



import os

from ch35t.parser import Parser
from ch35t.validator import Validator

class Chest():
    def __init__(self, chestfile, chests_dir="/tmp/chests"):
        self._chests_dir = chests_dir
        self._parser = Parser(chestfile)
        self.hint = self._parser.hint()
        self.payload = self._parser.payload()
        self.label = self._parser.label()
        self._validator = Validator()
        self._ctx = {
            "output_dir": chests_dir
        }
        
        if not os.path.exists(chests_dir):
            print(f"[i] Creating chests directory {chests_dir}")
            os.mkdir(chests_dir)

        # set context for hint and payload
        # Hint and payload might be null if there is a parsing
        # error OR if the parser has been explicitly initialized
        # with no JSON, for later manual loading (we should make
        # manual loading available to this class and the CLI too!)
        if self.hint is not None:
            self.hint.set_ctx(self._ctx)
        if self.payload is not None:
            self.payload.set_ctx(self._ctx)

    def unlock(self, key):
        # try to unlock the chest, if successful return True
        return self.payload.unlock(key)

    def json(self):
        return self._parser.json()

    def validate(self):
        self._validator.validate(self.json())

    def name(self):
        return self.label.name

    def author(self):
        return self.label.author



import json
import jsonschema
from pathlib import Path

SCHEMA_VERSION = "1.0.0"
SCHEMA_FILE_PATH = Path(__file__).resolve().parent / f"schema/{SCHEMA_VERSION}.json"

class ValidationError(Exception):
    def __init__(self, message):
        super().__init__(message)

class Validator():

    def __init__(self, schema_file=SCHEMA_FILE_PATH):
        with open(schema_file, "rt") as f:
            self._schema = json.load(f)

    def validate(self, json):
        try:
            jsonschema.validate(schema=self._schema, instance=json)
        except jsonschema.ValidationError as e:
            raise ValidationError(e.message)



import base64

class InvalidDataException(Exception):
    pass

def text_decode(data):
    '''
        If data is in a text-encoded format return the decoded
        version, otherwise return data.

        NOTE: currently supporting base64 only
    '''
    if data.startswith("base64,"):
        split_data = data.split(',')

        # we expect split_data to be exactly 2, as 
        # b64-encoded data does not contain commas
        if len(split_data) != 2:
            raise InvalidDataException

        return base64.b64decode(split_data[1])

    else:
        return data
